@@ -138,13 +138,32 @@ build g = g (:) []
 --   ('splitInternal' d l) == l@.
 splitInternal :: Delimiter a -> [a] -> SplitList a
 splitInternal _ [] = []
-splitInternal d xxs@(x:xs) = case matchDelim d xxs of
-                               -- special case for blank delimiter
-                               Just ([], (r:rs)) -> Delim [] : Text [r] : splitInternal d rs
-                               Just (match,rest) -> Delim match : splitInternal d rest
-                               _                 -> x `consText` splitInternal d xs
-  where consText z (Text c : ys) = Text (z:c) : ys
-        consText z ys             = Text [z] : ys
+splitInternal d xxs
+  | null xs   = toSplitList match
+  | otherwise = Text xs : toSplitList match
+ where
+  (xs,match) = breakDelim d xxs
+
+  toSplitList Nothing             = []
+  toSplitList (Just ([],r:rs))    = Delim [] : Text [r] : splitInternal d rs
+  toSplitList (Just (delim,rest)) = Delim delim : splitInternal d rest
+
+breakDelim :: Delimiter a -> [a] -> ([a],Maybe ([a],[a]))
+breakDelim d@(DelimEltPred p) (x:xs)
+  | p x       = ([],Just ([x],xs))
+  | otherwise = let (ys,match) = breakDelim d xs in (x:ys,match)
+breakDelim (DelimEltPred _) []  = ([],Nothing)
+breakDelim (DelimSublist []) xs = ([],Just ([],xs))
+breakDelim (DelimSublist _)  [] = ([],Nothing)
+breakDelim (DelimSublist ds) xxs@(x:xs) =
+  case matchSublist ds xxs of
+      Nothing   -> let (ys,match) = breakDelim (DelimSublist ds) xs in (x:ys,match)
+      Just rest -> ([],Just (ds,rest))
+
+matchSublist :: Eq a => [a] -> [a] -> Maybe [a]
+matchSublist [] xs = Just xs
+matchSublist _  [] = Nothing
+matchSublist (d:ds) (x:xs) = if d==x then matchSublist ds xs else Nothing
 
 -- | Given a split list in the internal tagged representation, produce
 --   a new internal tagged representation corresponding to the final
@@ -215,11 +234,12 @@ dropInitial _ l = l
 
 -- | Drop a final blank chunk according to the given 'EndPolicy'.
 dropFinal :: EndPolicy -> SplitList a -> SplitList a
-dropFinal _ [] = []
-dropFinal DropBlank l = case last l of
-                          Text [] -> init l
-                          _ -> l
-dropFinal _ l = l
+dropFinal _         [] = []
+dropFinal DropBlank l  = dropFinal' l
+  where dropFinal' []        = []
+        dropFinal' [Text []] = []
+        dropFinal' (x:xs)    = x:dropFinal' xs
+dropFinal _         l  = l
 
 -- * Combinators
 
@@ -328,7 +348,7 @@ dropBlanks = dropInitBlank . dropFinalBlank . condense
 --   Equivalent to @'dropInitBlank' . 'keepDelimsL' . 'onSublist'@.
 --   For example:
 --
--- > split (startsWith "app") "applyappicativeapplaudapproachapple" == ["apply","appicative","applaud","approach","apple"]
+-- > split (startsWith "app") "applyapplicativeapplaudapproachapple" == ["apply","applicative","applaud","approach","apple"]
 startsWith :: Eq a => [a] -> Splitter a
 startsWith = dropInitBlank . keepDelimsL . onSublist
 
@@ -455,6 +475,7 @@ linesBy = split . dropFinalBlank . dropDelims . whenElt
 --   whenever @n@ evenly divides the length of @xs@.
 splitEvery :: Int -> [e] -> [[e]]
 splitEvery i ls = map (take i) (build (splitter ls)) where
+  splitter :: [e] -> ([e] -> a -> a) -> a -> a
   splitter [] _ n = n
   splitter l c n  = l `c` splitter (drop i l) c n
 
@@ -472,7 +493,28 @@ chunk = splitEvery
 --   is total.
 splitPlaces :: Integral a => [a] -> [e] -> [[e]]
 splitPlaces is ys = build (splitPlacer is ys) where
+  splitPlacer :: forall i b t. Integral i
+              => [i] -> [b] -> ([b] -> t -> t) -> t -> t
   splitPlacer [] _ _ n      = n
   splitPlacer _ [] _ n      = n
   splitPlacer (l:ls) xs c n = let (x1, x2) = genericSplitAt l xs
                               in  x1 `c` splitPlacer ls x2 c n
+
+-- | A useful recursion pattern for processing a list to produce a new
+--   list, often used for \"chopping\" up the input list.  Typically
+--   chop is called with some function that will consume an initial
+--   prefix of the list and produce a value and the rest of the list.
+--
+--   For example, many common Prelude functions can be implemented in
+--   terms of @chop@:
+--
+-- > group :: (Eq a) => [a] -> [[a]]
+-- > group = chop (\ xs@(x:_) -> span (==x) xs)
+-- >
+-- > words :: String -> [String]
+-- > words = filter (not . null) . chop (span (not . isSpace) . dropWhile isSpace)
+
+chop :: ([a] -> (b, [a])) -> [a] -> [b]
+chop _ [] = []
+chop f as = b : chop f as'
+  where (b, as') = f as
